""" test ast utility functions """
import pytest
from clingo.ast import AST, ASTType, ComparisonOperator, parse_string

from ngo.normalize import postprocess, preprocess
from ngo.utils.ast import (
    Predicate,
    SignedPredicate,
    TranslationMap,
    collect_binding_information_body,
    collect_binding_information_head,
    compare,
    global_vars_inside_body,
    headderivable_predicates,
    potentially_unifying,
    replace_simple_assignments,
)


@pytest.mark.parametrize(
    "lhs, rhs",
    [
        ("X", "Y"),
        ("f(X)", "Y"),
        ("f(X)", "f(Y)"),
        ("|5|", "3-X"),
        ("-f(X)", "X"),
        ("-X", "-f(2)"),
        ("-X", "Y-B"),
    ],
)
def test_potentially_unifying_pos(lhs: str, rhs: str) -> None:
    """check potentially unifying patterns"""
    terms = []

    def assign(show: AST) -> None:
        if show.ast_type == ASTType.ShowTerm:
            terms.append(show.term)

    parse_string(f"#show {lhs}.", assign)
    parse_string(f"#show {rhs}.", assign)
    assert potentially_unifying(terms[0], terms[1])
    assert potentially_unifying(terms[1], terms[0])


@pytest.mark.parametrize(
    "lhs, rhs",
    [
        ("f(X)", "-X"),
        ("f(X)", "1"),
        ("a", "b"),
        ("f(X,Y)", "f(X,Y,Z)"),
        ("f(X,1)", "f(X,f(Y))"),
    ],
)
def test_potentially_unifying_neg(lhs: str, rhs: str) -> None:
    """check non unifying patterns"""
    terms = []

    def assign(show: AST) -> None:
        if show.ast_type == ASTType.ShowTerm:
            terms.append(show.term)

    parse_string(f"#show {(lhs)}.", assign)
    parse_string(f"#show {(rhs)}.", assign)
    assert not potentially_unifying(terms[0], terms[1])
    assert not potentially_unifying(terms[1], terms[0])


@pytest.mark.parametrize(
    "prg, head_preds",
    [
        (
            """
    a(1,X,_) :- b(X).
    """,
            [SignedPredicate(0, Predicate("a", 3))],
        ),
        (
            """
    {a(1,X) : c(42)} :- b(X).
    """,
            [SignedPredicate(0, Predicate("a", 2))],
        ),
        (
            """
    1 = #sum {1,a : a(X)} :- b(X).
    """,
            [SignedPredicate(0, Predicate("a", 1))],
        ),
    ],
)
def test_headderivable_predicates(prg: str, head_preds: list[SignedPredicate]) -> None:
    """test minmax aggregates on whole programs"""
    ast: list[AST] = []
    parse_string(prg, ast.append)
    assert sorted(head_preds) == sorted(headderivable_predicates(ast[1]))


@pytest.mark.parametrize(
    "prg, bound_vars, unbound_vars",
    [
        (
            """
:- b(Y), c(Y) : d(Z,W), not e(Z), not f(U).
""",
            ["Y"],
            ["U"],
        ),
        (
            """
:- Z = #sum {X,Y,W,V : b(VV)}.
    """,
            ["Z"],
            ["X", "Y", "W", "V"],
        ),
        (
            """
:- Z = #sum {X,Y,W,V : b(VV)} = ZZ.
    """,
            ["Z", "ZZ"],
            ["X", "Y", "W", "V"],
        ),
        (
            """
:- Z = #sum {X,Y,W,V : b(VV)} > ZZ.
    """,
            ["Z"],
            ["X", "Y", "W", "V", "ZZ"],
        ),
        (
            """
:- Z != #sum {X,Y,W,V : b(VV)}.
    """,
            [],
            ["X", "Y", "W", "V", "Z"],
        ),
        (
            """
:- X = Y, not Z = Y.
""",
            [],
            ["X", "Y", "Z"],
        ),
        (
            """
:- foo(X+Y,Z).
""",
            ["Z"],
            ["X", "Y"],
        ),
        (
            """
:- b(Y), |X|=Y.
""",
            ["Y"],
            ["X"],
        ),
        (
            """
:- at2(XX,YY,Z); Z = ((S+1)..(S+2)).
""",
            ["XX", "YY", "Z"],
            ["S"],
        ),
        (
            """
:- at(X); X-W = 0.
""",
            ["X"],
            ["W"],
        ),
        (
            """
:- b(Y), c(Y,YY,Z) : d(Z,W), not e(Z), not f(U).
""",
            ["Y"],
            ["U", "YY"],
        ),
        (
            """
:- a(X,Y), not b(Z,W), (X,Y) = (Z,W).
""",
            ["X", "Y", "Z", "W"],
            [],
        ),
        (
            """
:- foo(Y), not Z = Y.
""",
            ["Y"],
            ["Z"],
        ),
    ],
)
def test_binding_variables_body(prg: str, bound_vars: list[str], unbound_vars: list[str]) -> None:
    """test finding bound variables in rules"""
    ast: list[AST] = []
    parse_string(prg, ast.append)
    bound, unbound = collect_binding_information_body(ast[1].body)
    assert set(bound_vars) == set(x.name for x in bound)
    assert set(unbound_vars) == set(x.name for x in unbound)


@pytest.mark.parametrize(
    "prg, need_bound, no_bound_needed",
    [
        (
            """
1 <= { order(T,S): S = (1..Nr), task_nr(Nr) } <= 1 :- task(T).
""",
            [],
            ["T", "S", "Nr"],
        ),
        (
            """
1 <= { order(T,S): S = (1..Nr), task_nr(Nr) } <= 1.
""",
            ["T"],
            ["S", "Nr"],
        ),
        (
            """
1 <= #sum{ T,S: order(T,S): S = (1..Nr), task_nr(Nr) } <= 1 :- task(T).
""",
            [],
            ["T", "S", "Nr"],
        ),
        (
            """
1 <= #sum{ T,S: order(T,S): S = (1..Nr), task_nr(Nr) } <= 1.
""",
            ["T"],
            ["S", "Nr"],
        ),
        (
            """
S <= 5 :- dom(S).
""",
            [],
            ["S"],
        ),
        (
            """
S <= 5.
""",
            ["S"],
            [],
        ),
        (
            """
p(A); p(B) : dom(B).
""",
            ["A"],
            ["B"],
        ),
        (
            """
1 = { sudoku(X,Y,N): N = (1..(S*S)) }.
""",
            ["S", "N", "X", "Y"],
            [],
        ),
    ],
)
def test_binding_variables_head(prg: str, need_bound: list[str], no_bound_needed: list[str]) -> None:
    """test finding bound variables in rules"""
    ast: list[AST] = []
    parse_string(prg, ast.append)
    bound, unbound = collect_binding_information_head(ast[1].head, ast[1].body)
    assert set(need_bound) == set(x.name for x in bound)
    assert set(no_bound_needed) == set(x.name for x in unbound)


@pytest.mark.parametrize(
    "prg, globals_",
    [
        (
            """
:- b(Y), c(Y) : d(Z,W), not e(Z), not f(U).
""",
            ["Y", "U"],
        ),
        (
            """
:- Z = #sum {X,Y,W,V : b(VV)}.
    """,
            ["Z", "X", "Y", "W", "V"],
        ),
        (
            """
:- Z = #sum {X,Y,W,V : b(VV)} != I.
    """,
            ["Z", "X", "Y", "W", "V", "I"],
        ),
        (
            """
:- Z != #sum {X,Y,W,V : b(VV)}.
    """,
            ["X", "Y", "W", "V", "Z"],
        ),
        (
            """
:- Z != #sum {X,Y,W,V : not b(VV)}.
    """,
            ["X", "Y", "W", "V", "Z", "VV"],
        ),
        (
            """
:- X = Y, not Z = Y.
""",
            ["X", "Y", "Z"],
        ),
        (
            """
:- foo(X+Y,Z).
""",
            ["Z", "X", "Y"],
        ),
        (
            """
:- b(Y), |X|=Y.
""",
            ["Y", "X"],
        ),
        (
            """
:- at2(XX,YY,Z); Z = ((S+1)..(S+2)).
""",
            ["XX", "YY", "Z", "S"],
        ),
        (
            """
:- at(X); X-W = 0.
""",
            ["X", "W"],
        ),
        (
            """
:- b(Y), c(Y,YY,Z) : d(Z,W), not e(Z), not f(U).
""",
            ["Y", "YY", "U"],
        ),
    ],
)
def test_global_vars_body(prg: str, globals_: list[str]) -> None:
    """test minmax aggregates on whole programs"""
    ast: list[AST] = []
    parse_string(prg, ast.append)
    assert set(globals_) == set(map(lambda x: x.name, global_vars_inside_body(ast[1].body)))


@pytest.mark.parametrize(
    "input_, output",
    [
        (
            """
a(X,Y) :- b(X), c(Y), X = Y.
""",
            """a(X,X) :- b(X); c(X).""",
        ),
        (
            """
a(X,Y) :- b(X), c(Y), not X != Y.
""",
            """a(X,X) :- b(X); c(X).""",
        ),
        (
            """
a(Z,Y) :- b(X), c(Y), X = Y, Y = Z.
""",
            """a(X,X) :- b(X); c(X).""",
        ),
        (
            """
a(Z,Y) :- b(X), c(Y), X = Y = Z > 2.
""",
            """a(X,X) :- b(X); c(X); X > 2.""",
        ),
        (
            """
a(X,Y) :- b(X), c(Y), X = Y, #sum {A : dom(A,B), B = A}.
""",
            """a(X,X) :- b(X); c(X); #sum { A: dom(A,A) }.""",
        ),
        (
            """
a(X,Y) :- b(X), c(Y), X = Y, #sum {A : dom(A,B), B = A; A,B : foo(A,B)}.
""",
            """a(X,X) :- b(X); c(X); #sum { A: dom(A,A); A,B: foo(A,B) }.""",
        ),
        (
            """
a(X,Y) :- bar(A,B), b(X), c(Y), X = Y, #sum {A : dom(A,B), B = A; A,B : foo(A,B)}.
""",
            """a(X,X) :- bar(A,B); b(X); c(X); #sum { A: dom(A,A); A,B: foo(A,B) }.""",
        ),
        (
            """
#show.
""",
            """#show.""",
        ),
        (
            """
#minimize {T,U : all_finish(T, U), T=U}.
""",
            """:~ all_finish(T,T). [T@0,T]""",
        ),
        (
            """
working(J2,1,D) :- perm(J1,P1); working(J1,1,D1); perm(J2,P2);\
 duration(J2,1,D2); P2 = (P1+1); D = (D2+D1); D <= S; sum_duration(S).
""",
            """working(J2,1,(D2+D1)) :- perm(J1,P1); working(J1,1,D1);\
 perm(J2,(P1+1)); duration(J2,1,D2); (D2+D1) <= S; sum_duration(S).""",
        ),
    ],
)
def test_replace_simple_assignments(input_: str, output: str) -> None:
    """test replacing assignments in rules"""
    ast: list[AST] = []
    parse_string(input_, ast.append)
    ast = preprocess(ast)
    ast[1] = replace_simple_assignments(ast[1])
    ast = postprocess(ast)
    assert output == str(ast[1])


def test_translation_mapping() -> None:
    """test the argument order translation class"""
    # NOTE: isn't there something already in python that can reorder things ?
    trans = TranslationMap(Predicate("a", 3), Predicate("b", 4), (2, 1, 0))
    assert trans.translate_parameters(["X", "Y", "Z"]) == ["Z", "Y", "X"]
    trans = TranslationMap(Predicate("a", 3), Predicate("b", 3), (2, 1))
    assert trans.translate_parameters(["X", "Y", "Z"]) == [None, "Y", "X"]
    trans = TranslationMap(Predicate("a", 3), Predicate("b", 3), (None, 2, 1))
    assert trans.translate_parameters(["X", "Y", "Z"]) == [None, "Z", "Y"]


def test_compare() -> None:
    """test comparison"""
    assert compare(1, ComparisonOperator.Equal, 1)
    assert compare(1, ComparisonOperator.NotEqual, 2)
    assert compare(1, ComparisonOperator.LessEqual, 2)
    assert compare(1, ComparisonOperator.LessThan, 2)
    assert compare(1, ComparisonOperator.GreaterEqual, 1)
    assert compare(2, ComparisonOperator.GreaterThan, 1)

    assert not compare(1, ComparisonOperator.Equal, 2)
    assert not compare(1, ComparisonOperator.NotEqual, 1)
    assert not compare(1, ComparisonOperator.LessEqual, 0)
    assert not compare(1, ComparisonOperator.LessThan, 1)
    assert not compare(1, ComparisonOperator.GreaterEqual, 2)
    assert not compare(1, ComparisonOperator.GreaterThan, 1)



""" test translation of sum aggregates using chaining """
import pytest
from clingo.ast import AST, parse_string

from ngo.normalize import postprocess, preprocess
from ngo.sum_aggregates import SumAggregator
from ngo.utils.ast import AnnotatedPredicate, Predicate


@pytest.mark.parametrize(
    "prg, at_most_one, at_least_one",
    [
        (
            """
{ shift(D,L) : pshift(_,L) } 1 :- day(D).
         """,
            [
                AnnotatedPredicate(Predicate("shift", 2), (1,)),
            ],
            [],
        ),
        (
            """
#sum { 1 : shift(D,L) : pshift(_,L) } 1 :- day(D).
         """,
            [
                AnnotatedPredicate(Predicate("shift", 2), (1,)),
            ],
            [],
        ),
        (
            """
 3 <= #sum { 1 : shift(D,L) : pshift(_,L) } <= 0 :- day(D).
         """,
            [
                AnnotatedPredicate(Predicate("shift", 2), (1,)),
            ],
            [
                AnnotatedPredicate(Predicate("shift", 2), (1,)),
            ],
        ),
        (
            """
 3 < #sum { 1 : shift(D,L) : pshift(_,L) } < 2 :- day(D).
         """,
            [
                AnnotatedPredicate(Predicate("shift", 2), (1,)),
            ],
            [
                AnnotatedPredicate(Predicate("shift", 2), (1,)),
            ],
        ),
        (
            """
 #sum { 1 : shift(D,L) : pshift(_,L) }:- day(D).
         """,
            [],
            [],
        ),
        (
            """
a :- b.
         """,
            [],
            [],
        ),
        (  # L could be negative
            """
 #sum { L : shift(D,L) : pshift(_,L) } 1:- day(D).
         """,
            [],
            [],
        ),
        (
            """
#sum { 1,D,L : not not shift(D,L) : pshift(_,L);  1,b,D,L : shift(D,L) : pshift(_,L) } 1 :- day(D).
         """,
            [
                AnnotatedPredicate(Predicate("shift", 2), (1,)),
            ],
            [],
        ),
        (
            """
#sum { 1,D,L : L> 42 : pshift(_,L);  1,b,D,L : shift(D,L) : pshift(_,L) } 1 :- day(D).
         """,
            [
                AnnotatedPredicate(Predicate("shift", 2), (1,)),
            ],
            [],
        ),
        (
            """
#sum { 1,D,L : #true : pshift(_,L);  1,b,D,L : shift(D,L) : pshift(_,L) } 1 :- day(D).
         """,
            [
                AnnotatedPredicate(Predicate("shift", 2), (1,)),
            ],
            [],
        ),
    ],
)
def test_sum_aggregates_bound_detection(
    prg: str, at_most_one: list[AnnotatedPredicate], at_least_one: list[AnnotatedPredicate]
) -> None:
    """test sum aggregates on whole programs"""
    ast: list[AST] = []  # pylint: disable=duplicate-code
    parse_string(prg, ast.append)
    ast = preprocess(ast)
    mma = SumAggregator(ast, [])
    assert sorted(mma.at_most_one_predicates()) == at_most_one
    assert sorted(mma.at_least_one_predicates()) == at_least_one


@pytest.mark.parametrize(
    "prg, converted_prg",
    [
        (
            """
{ shift(D,L) : pshift(_,L) } 1 :- day(D).
         """,
            """#program base.
1 >= { shift(D,L): pshift(_,L) } :- day(D).""",
        ),
        (
            """
{ shift(D,L) : pshift(D,L) } 1 :- day(D).
a(X) :- X = #sum {L,D : shift(D,L)}.
         """,
            """#program base.
1 >= { shift(D,L): pshift(D,L) } :- day(D).
__dom_shift(D,L) :- pshift(D,L); day(D).
__min_1_1__dom_shift(G0,X) :- X = #min { L: __dom_shift(G0,L) }; __dom_shift(G0,_).
__max_1_1__dom_shift(G0,X) :- X = #max { L: __dom_shift(G0,L) }; __dom_shift(G0,_).
__next_1_1__dom_shift(G0,P,N) :- __min_1_1__dom_shift(G0,P); __dom_shift(G0,N); N > P;\
 not __dom_shift(G0,B): __dom_shift(G0,B), P < B < N.
__next_1_1__dom_shift(G0,P,N) :- __next_1_1__dom_shift(G0,_,P); __dom_shift(G0,N); N > P;\
 not __dom_shift(G0,B): __dom_shift(G0,B), P < B < N.
__chain_1_1__max___dom_shift(G0,P) :- shift(G0,P).
__chain_1_1__max___dom_shift(G0,P) :- __chain_1_1__max___dom_shift(G0,N); __next_1_1__dom_shift(G0,P,N).
a(X) :- X = #sum { (L-__PREV),D,__next_1_1__dom_shift(D,__PREV,L):\
 __chain_1_1__max___dom_shift(D,L), __next_1_1__dom_shift(D,__PREV,L);\
 L,D,__next_1_1__dom_shift(D,L): __chain_1_1__max___dom_shift(D,L), not __next_1_1__dom_shift(D,_,L) }.""",
        ),
        (
            """
{ shift(D,L) : pshift(D,L) } 1 :- day(D).
a(X) :- X = #sum {L,D : shift(D,L)}, day(D).
         """,
            """#program base.
1 >= { shift(D,L): pshift(D,L) } :- day(D).
__dom_shift(D,L) :- pshift(D,L); day(D).
__min_1_1__dom_shift(G0,X) :- X = #min { L: __dom_shift(G0,L) }; __dom_shift(G0,_).
__max_1_1__dom_shift(G0,X) :- X = #max { L: __dom_shift(G0,L) }; __dom_shift(G0,_).
__next_1_1__dom_shift(G0,P,N) :- __min_1_1__dom_shift(G0,P); __dom_shift(G0,N); N > P;\
 not __dom_shift(G0,B): __dom_shift(G0,B), P < B < N.
__next_1_1__dom_shift(G0,P,N) :- __next_1_1__dom_shift(G0,_,P); __dom_shift(G0,N); N > P;\
 not __dom_shift(G0,B): __dom_shift(G0,B), P < B < N.
__chain_1_1__max___dom_shift(G0,P) :- shift(G0,P).
__chain_1_1__max___dom_shift(G0,P) :- __chain_1_1__max___dom_shift(G0,N); __next_1_1__dom_shift(G0,P,N).
a(X) :- X = #sum { (L-__PREV),D,__next_1_1__dom_shift(D,__PREV,L):\
 __chain_1_1__max___dom_shift(D,L), __next_1_1__dom_shift(D,__PREV,L);\
 L,D,__next_1_1__dom_shift(D,L): __chain_1_1__max___dom_shift(D,L), not __next_1_1__dom_shift(D,_,L) }; day(D).""",
        ),
        (
            """
{ shift(D,L) : pshift(D,L) } 1 :- day(D).
#minimize {L,D : shift(D,L)}.
         """,
            """#program base.
1 >= { shift(D,L): pshift(D,L) } :- day(D).
__dom_shift(D,L) :- pshift(D,L); day(D).
__min_1_1__dom_shift(G0,X) :- X = #min { L: __dom_shift(G0,L) }; __dom_shift(G0,_).
__max_1_1__dom_shift(G0,X) :- X = #max { L: __dom_shift(G0,L) }; __dom_shift(G0,_).
__next_1_1__dom_shift(G0,P,N) :- __min_1_1__dom_shift(G0,P); __dom_shift(G0,N); N > P;\
 not __dom_shift(G0,B): __dom_shift(G0,B), P < B < N.
__next_1_1__dom_shift(G0,P,N) :- __next_1_1__dom_shift(G0,_,P); __dom_shift(G0,N); N > P;\
 not __dom_shift(G0,B): __dom_shift(G0,B), P < B < N.
__chain_1_1__max___dom_shift(G0,P) :- shift(G0,P).
__chain_1_1__max___dom_shift(G0,P) :- __chain_1_1__max___dom_shift(G0,N); __next_1_1__dom_shift(G0,P,N).
:~ __chain_1_1__max___dom_shift(D,L);\
 __next_1_1__dom_shift(D,__PREV,L). [(L-__PREV)@0,D,__next_1_1__dom_shift(D,__PREV,L)]
:~ __chain_1_1__max___dom_shift(D,L); not __next_1_1__dom_shift(D,_,L). [L@0,D,__next_1_1__dom_shift(D,L)]""",
        ),
        (
            """
{ shift(D,L) : pshift(D,L) } 1 :- day(D).
#maximize {L,D : shift(D,L)}.
         """,
            """#program base.
1 >= { shift(D,L): pshift(D,L) } :- day(D).
__dom_shift(D,L) :- pshift(D,L); day(D).
__min_1_1__dom_shift(G0,X) :- X = #min { L: __dom_shift(G0,L) }; __dom_shift(G0,_).
__max_1_1__dom_shift(G0,X) :- X = #max { L: __dom_shift(G0,L) }; __dom_shift(G0,_).
__next_1_1__dom_shift(G0,P,N) :- __min_1_1__dom_shift(G0,P); __dom_shift(G0,N); N > P;\
 not __dom_shift(G0,B): __dom_shift(G0,B), P < B < N.
__next_1_1__dom_shift(G0,P,N) :- __next_1_1__dom_shift(G0,_,P); __dom_shift(G0,N); N > P;\
 not __dom_shift(G0,B): __dom_shift(G0,B), P < B < N.
__chain_1_1__max___dom_shift(G0,P) :- shift(G0,P).
__chain_1_1__max___dom_shift(G0,P) :- __chain_1_1__max___dom_shift(G0,N); __next_1_1__dom_shift(G0,P,N).
:~ __chain_1_1__max___dom_shift(D,L);\
 __next_1_1__dom_shift(D,__PREV,L). [-(L-__PREV)@0,D,__next_1_1__dom_shift(D,__PREV,L)]
:~ __chain_1_1__max___dom_shift(D,L); not __next_1_1__dom_shift(D,_,L). [-L@0,D,__next_1_1__dom_shift(D,L)]""",
        ),
        (
            """
{ shift(D,L) : pshift(D,L) } 1 :- day(D).
:~ shift(D,L) : pshift(D,L). [L,D]""",
            """#program base.
1 >= { shift(D,L): pshift(D,L) } :- day(D).
:~ shift(D,L): pshift(D,L). [L@0,D]""",
        ),
        (
            """
{ shift(D,L) : pshift(D,L) } 1 :- day(D).
:~ shift(D,L): a. [L,D]""",
            """#program base.
1 >= { shift(D,L): pshift(D,L) } :- day(D).
:~ shift(D,L): a. [L@0,D]""",
        ),
        (
            """
{ shift(D,L) : pshift(D,L) } 1 :- day(D).
#minimize {L,D : #true, shift(D,L)}.
         """,
            """#program base.
1 >= { shift(D,L): pshift(D,L) } :- day(D).
__dom_shift(D,L) :- pshift(D,L); day(D).
__min_1_1__dom_shift(G0,X) :- X = #min { L: __dom_shift(G0,L) }; __dom_shift(G0,_).
__max_1_1__dom_shift(G0,X) :- X = #max { L: __dom_shift(G0,L) }; __dom_shift(G0,_).
__next_1_1__dom_shift(G0,P,N) :- __min_1_1__dom_shift(G0,P); __dom_shift(G0,N); N > P;\
 not __dom_shift(G0,B): __dom_shift(G0,B), P < B < N.
__next_1_1__dom_shift(G0,P,N) :- __next_1_1__dom_shift(G0,_,P); __dom_shift(G0,N); N > P;\
 not __dom_shift(G0,B): __dom_shift(G0,B), P < B < N.
__chain_1_1__max___dom_shift(G0,P) :- shift(G0,P).
__chain_1_1__max___dom_shift(G0,P) :- __chain_1_1__max___dom_shift(G0,N); __next_1_1__dom_shift(G0,P,N).
:~ #true; __chain_1_1__max___dom_shift(D,L);\
 __next_1_1__dom_shift(D,__PREV,L). [(L-__PREV)@0,D,__next_1_1__dom_shift(D,__PREV,L)]
:~ #true; __chain_1_1__max___dom_shift(D,L); not __next_1_1__dom_shift(D,_,L). [L@0,D,__next_1_1__dom_shift(D,L)]""",
        ),
        (
            """
{ shift(D,L,foo) : pshift(D,L) } 1 :- day(D).
#minimize {L,D : shift(D,L,_)}.
         """,
            """#program base.
1 >= { shift(D,L,foo): pshift(D,L) } :- day(D).
__dom_shift(D,L,foo) :- pshift(D,L); day(D).
__min_1_1__dom_shift(G0,G2,X) :- X = #min { L: __dom_shift(G0,L,G2) }; __dom_shift(G0,_,G2).
__max_1_1__dom_shift(G0,G2,X) :- X = #max { L: __dom_shift(G0,L,G2) }; __dom_shift(G0,_,G2).
__next_1_1__dom_shift(G0,G2,P,N) :- __min_1_1__dom_shift(G0,G2,P); __dom_shift(G0,N,G2); N > P;\
 not __dom_shift(G0,B,G2): __dom_shift(G0,B,G2), P < B < N.
__next_1_1__dom_shift(G0,G2,P,N) :- __next_1_1__dom_shift(G0,G2,_,P); __dom_shift(G0,N,G2); N > P;\
 not __dom_shift(G0,B,G2): __dom_shift(G0,B,G2), P < B < N.
__chain_1_1__max___dom_shift(G0,G2,P) :- shift(G0,P,G2).
__chain_1_1__max___dom_shift(G0,G2,P) :- __chain_1_1__max___dom_shift(G0,G2,N); __next_1_1__dom_shift(G0,G2,P,N).
:~ __chain_1_1__max___dom_shift(D,_,L);\
 __next_1_1__dom_shift(D,_,__PREV,L). [(L-__PREV)@0,D,__next_1_1__dom_shift(D,none,__PREV,L)]
:~ __chain_1_1__max___dom_shift(D,_,L); not __next_1_1__dom_shift(D,_,_,L). [L@0,D,__next_1_1__dom_shift(D,none,L)]""",
        ),
        (
            """
{ shift(D,L,X) : pshift(D,L,X) } 1 :- day(D).
#minimize {L,D : shift(D,L,_)}.
         """,
            """#program base.
1 >= { shift(D,L,X): pshift(D,L,X) } :- day(D).
__dom_shift(D,L,X) :- pshift(D,L,X); day(D).
__min_1_2_1__dom_shift(G0,X) :- X = #min { L: __dom_shift(G0,L,_) }; __dom_shift(G0,_,_).
__max_1_2_1__dom_shift(G0,X) :- X = #max { L: __dom_shift(G0,L,_) }; __dom_shift(G0,_,_).
__next_1_2_1__dom_shift(G0,P,N) :- __min_1_2_1__dom_shift(G0,P); __dom_shift(G0,N,_); N > P;\
 not __dom_shift(G0,B,_): __dom_shift(G0,B,_), P < B < N.
__next_1_2_1__dom_shift(G0,P,N) :- __next_1_2_1__dom_shift(G0,_,P); __dom_shift(G0,N,_); N > P;\
 not __dom_shift(G0,B,_): __dom_shift(G0,B,_), P < B < N.
__chain_1_2_1__max___dom_shift(G0,P) :- shift(G0,P,_).
__chain_1_2_1__max___dom_shift(G0,P) :- __chain_1_2_1__max___dom_shift(G0,N); __next_1_2_1__dom_shift(G0,P,N).
:~ __chain_1_2_1__max___dom_shift(D,L);\
 __next_1_2_1__dom_shift(D,__PREV,L). [(L-__PREV)@0,D,__next_1_2_1__dom_shift(D,__PREV,L)]
:~ __chain_1_2_1__max___dom_shift(D,L); not __next_1_2_1__dom_shift(D,_,L). [L@0,D,__next_1_2_1__dom_shift(D,L)]""",
        ),
        (
            """
{ shift(D,L,X) : pshift(D,L,X) } 1 :- day(D).
#minimize {L,D : shift(D,L,X)}.
         """,
            """#program base.
1 >= { shift(D,L,X): pshift(D,L,X) } :- day(D).
:~ shift(D,L,X). [L@0,D]""",
        ),
        (
            """
{ shift(D,L,X) : pshift(D,L,X) } 1 :- day(D).
#minimize {L*2,D : shift(D,L,_)}.
         """,
            """#program base.
1 >= { shift(D,L,X): pshift(D,L,X) } :- day(D).
:~ shift(D,L,_). [(L*2)@0,D]""",
        ),
        (
            """
{ shift(D,L) : pshift(D,L) } 1 :- day(D).
a(X) :- X = #sum {L,D : shift(D,L)}.
         """,
            """#program base.
1 >= { shift(D,L): pshift(D,L) } :- day(D).
__dom_shift(D,L) :- pshift(D,L); day(D).
__min_1_1__dom_shift(G0,X) :- X = #min { L: __dom_shift(G0,L) }; __dom_shift(G0,_).
__max_1_1__dom_shift(G0,X) :- X = #max { L: __dom_shift(G0,L) }; __dom_shift(G0,_).
__next_1_1__dom_shift(G0,P,N) :- __min_1_1__dom_shift(G0,P); __dom_shift(G0,N); N > P;\
 not __dom_shift(G0,B): __dom_shift(G0,B), P < B < N.
__next_1_1__dom_shift(G0,P,N) :- __next_1_1__dom_shift(G0,_,P); __dom_shift(G0,N); N > P;\
 not __dom_shift(G0,B): __dom_shift(G0,B), P < B < N.
__chain_1_1__max___dom_shift(G0,P) :- shift(G0,P).
__chain_1_1__max___dom_shift(G0,P) :- __chain_1_1__max___dom_shift(G0,N); __next_1_1__dom_shift(G0,P,N).
a(X) :- X = #sum { (L-__PREV),D,__next_1_1__dom_shift(D,__PREV,L):\
 __chain_1_1__max___dom_shift(D,L), __next_1_1__dom_shift(D,__PREV,L);\
 L,D,__next_1_1__dom_shift(D,L): __chain_1_1__max___dom_shift(D,L), not __next_1_1__dom_shift(D,_,L) }.""",
        ),
        (
            """
{ shift(D,L) : pshift(D,L) } 1 :- day(D).
a(X) :- X = #sum {L*2,D : shift(D,L)}.
         """,
            """#program base.
1 >= { shift(D,L): pshift(D,L) } :- day(D).
a(X) :- X = #sum { (L*2),D: shift(D,L) }.""",
        ),
        (
            """
:- day(D) : a.
         """,
            """#program base.
#false :- day(D): a.""",
        ),
        (
            """
{ shift(D,L) : pshift(D,L) } 1 :- day(D).
a(X) :- X = #sum { 1 }.
         """,
            """#program base.
1 >= { shift(D,L): pshift(D,L) } :- day(D).
a(X) :- X = #sum { 1 }.""",
        ),
        (
            """
{ shift(D,L) : pshift(D,L) } 1 :- day(D).
a(X) :- X = #sum { 1,extra; V: bla(V); Y,foo,bar }, some(Y).
         """,
            """#program base.
1 >= { shift(D,L): pshift(D,L) } :- day(D).
a(X) :- X = #sum { 1,extra; V: bla(V); Y,foo,bar }; some(Y).""",
        ),
        (  # non unique tuple problem
            """
{ shift(D,L) : pshift(D,L) } 1 :- day(D).
#minimize {L,D : shift(D,L)}.
#minimize {1,a : #true}.
         """,
            """#program base.
1 >= { shift(D,L): pshift(D,L) } :- day(D).
:~ shift(D,L). [L@0,D]
:~ #true. [1@0,a]""",
        ),
        (  # non unique tuple in sums
            """
{ shift(D,L) : pshift(D,L) } 1 :- day(D).
a(X) :- X = #sum {L,D : shift(D,L); 1,a : #true}.
         """,
            """#program base.
1 >= { shift(D,L): pshift(D,L) } :- day(D).
a(X) :- X = #sum { L,D: shift(D,L); 1,a: #true }.""",
        ),
        (  # non unique tuple in sums
            """
{ shift(D,L) : pshift(D,L) } 1 :- day(D).
a(X) :- X = #sum {L,D : shift(D,L); 1,a : #true}.
         """,
            """#program base.
1 >= { shift(D,L): pshift(D,L) } :- day(D).
a(X) :- X = #sum { L,D: shift(D,L); 1,a: #true }.""",
        ),
        (
            """
{ shift(D,L) : pshift(D,L) } 1 :- day(D).
:~ X = #sum {L,D : shift(D,L)}, p(X,Y). [Y@1]
         """,
            """#program base.
1 >= { shift(D,L): pshift(D,L) } :- day(D).
__dom_shift(D,L) :- pshift(D,L); day(D).
__min_1_1__dom_shift(G0,X) :- X = #min { L: __dom_shift(G0,L) }; __dom_shift(G0,_).
__max_1_1__dom_shift(G0,X) :- X = #max { L: __dom_shift(G0,L) }; __dom_shift(G0,_).
__next_1_1__dom_shift(G0,P,N) :- __min_1_1__dom_shift(G0,P); __dom_shift(G0,N); N > P;\
 not __dom_shift(G0,B): __dom_shift(G0,B), P < B < N.
__next_1_1__dom_shift(G0,P,N) :- __next_1_1__dom_shift(G0,_,P); __dom_shift(G0,N); N > P;\
 not __dom_shift(G0,B): __dom_shift(G0,B), P < B < N.
__chain_1_1__max___dom_shift(G0,P) :- shift(G0,P).
__chain_1_1__max___dom_shift(G0,P) :- __chain_1_1__max___dom_shift(G0,N); __next_1_1__dom_shift(G0,P,N).
:~ X = #sum { (L-__PREV),D,__next_1_1__dom_shift(D,__PREV,L):\
 __chain_1_1__max___dom_shift(D,L), __next_1_1__dom_shift(D,__PREV,L);\
 L,D,__next_1_1__dom_shift(D,L): __chain_1_1__max___dom_shift(D,L), not __next_1_1__dom_shift(D,_,L) };\
 p(X,Y). [Y@1]""",
        ),
    ],
)
def test_sum_aggregates_output(prg: str, converted_prg: str) -> None:
    """test sum aggregates on whole programs"""
    ast: list[AST] = []  # pylint: disable=duplicate-code
    parse_string(prg, ast.append)
    ast = preprocess(ast)
    mma = SumAggregator(ast, [])
    ast = mma.execute(ast)
    ast = postprocess(ast)
    output = "\n".join(map(str, ast))
    assert converted_prg == output



"""
 This module replaces bodys with aggregates of the form X = #max/min{}
 with an order encoding.
 Might also replace the usage of the resulting predicate with the order literals.
"""

import logging
from collections import defaultdict
from itertools import chain
from typing import Callable, Iterator, Optional

from clingo.ast import (
    AST,
    AggregateFunction,
    ASTType,
    BinaryOperation,
    BinaryOperator,
    BodyAggregate,
    BodyAggregateElement,
    Comparison,
    ComparisonOperator,
    ConditionalLiteral,
    Function,
    Guard,
    Literal,
    Location,
    Minimize,
    Rule,
    Sign,
    SymbolicAtom,
    SymbolicTerm,
    UnaryOperation,
    UnaryOperator,
    Variable,
)
from clingo.symbol import Infimum, Supremum

from ngo.dependency import DomainPredicates, RuleDependency
from ngo.utils.ast import (
    LOC,
    AggAnalytics,
    AnnotatedPredicate,
    Predicate,
    TranslationMap,
    collect_ast,
    global_vars_inside_body,
    is_predicate,
    loc2str,
    potentially_unifying_sequence,
    predicates,
    transform_ast,
)
from ngo.utils.globals import NEXT, PREV, UniqueNames, UniqueVariables

log = logging.getLogger(__name__)


def _characteristic_variables(term: AST) -> Iterator[AST]:
    """yield all characteristic variable names of the given Term,
    this means all Variables not occuring in arithmetics, bounds, etc...
    Assumes term is unpooled.
    """

    if term.ast_type in {ASTType.Variable, ASTType.SymbolicTerm}:
        yield from collect_ast(term, "Variable")
    elif term.ast_type == ASTType.Function:
        for i in term.arguments:
            yield from _characteristic_variables(i)


class MinMaxAggregator:
    """Translates some min/max aggregates into chains"""

    MinMaxPred = tuple[int, TranslationMap, int]

    def __init__(self, prg: list[AST], input_predicates: list[Predicate]):
        self.unique_names = UniqueNames(prg, input_predicates)
        self.rule_dependency = RuleDependency(prg)
        self.domain_predicates = DomainPredicates(self.unique_names, prg)
        # list of ({AggregateFunction.Max, AggregateFunction.Min}, Translation, index)
        #  where index is the position of the variable indicating the minimum/maximum

        self._minmax_preds: list[MinMaxAggregator.MinMaxPred] = []

    def _store_aggregate_head(
        self, function: int, head: AST, rest_vars: list[AST], max_var: AST, new_name: str
    ) -> None:
        """
        if this head represents the result of the aggregate
        store it in self._minmax_preds for other simplifications
        """

        if not (
            is_predicate(head)
            and len(self.rule_dependency.get_bodies(Predicate(head.atom.symbol.name, len(head.atom.symbol.arguments))))
            == 1  # only this head occurence
        ):
            return
        symbol = head.atom.symbol
        for arg in symbol.arguments:
            if arg.ast_type not in {ASTType.Variable, ASTType.SymbolicTerm}:
                return  # nocoverage

        mapping = [
            (rest_vars + [max_var]).index(arg) if arg in rest_vars + [max_var] else None for arg in symbol.arguments
        ]

        translation = TranslationMap(
            Predicate(symbol.name, len(symbol.arguments)),
            Predicate(new_name, len(rest_vars) + 1),
            mapping,
        )
        for i, arg in enumerate(symbol.arguments):
            if arg.ast_type == ASTType.Variable and arg.name == max_var.name:
                self._minmax_preds.append((function, translation, i))

    def _store_aggregate_for_minimize(self, function: int, rest_vars: list[AST], max_var: AST, new_name: str) -> None:
        """
        store it in self._minmax_preds for other simplifications
        """
        args = rest_vars + [max_var]

        for arg in args:
            if arg.ast_type not in {ASTType.Variable, ASTType.SymbolicTerm}:
                return  # nocoverage
        symbol = Function(LOC, new_name, args, False)

        mapping = [(args).index(arg) if arg in args else None for arg in args]

        translation = TranslationMap(
            Predicate(symbol.name, len(symbol.arguments)),
            Predicate(new_name, len(rest_vars) + 1),
            mapping,
        )
        for i, arg in enumerate(symbol.arguments):
            if arg.ast_type == ASTType.Variable and arg.name == max_var.name:
                self._minmax_preds.append((function, translation, i))

    def _translatable_element(self, elem: AST) -> bool:
        """check if this element of a minmax aggregate is currently worth translating
        return false if the element is static
        """

        def static(cond: AST) -> bool:
            if cond.ast_type == ASTType.Literal and cond.atom.ast_type == ASTType.SymbolicAtom:
                condition_pred = Predicate(
                    cond.atom.symbol.name,
                    len(cond.atom.symbol.arguments),
                )
                return self.domain_predicates.is_static(condition_pred)
            return True

        return not all(map(static, elem.condition))

    def _minmax_agg(self, rule: AST) -> Optional[AST]:
        """
        returns the first min/max aggregate in the rule body
        or None if none found
        """
        for blit in rule.body:
            assert isinstance(blit, AST)
            if blit.ast_type == ASTType.Literal:
                atom = blit.atom
                if atom.ast_type == ASTType.BodyAggregate and atom.function in {
                    AggregateFunction.Max,
                    AggregateFunction.Min,
                }:
                    return blit
        return None

    def _create_aggregate_replacement(
        self,
        agg: AST,
        elem: AST,
        rest_vars: list[AST],
        new_predicate: Predicate,
        lits_with_vars: list[AST],
    ) -> list[AST]:
        """return a list of rules as a replacement of this aggregate element"""
        # 2. create dom and next_ predicates for it, and then use it to
        # create chain with elem.condition + lits_with_vars
        # __chain__max_0_0_11(P,V) :- skill(P,ID,V); person(P).
        # __chain__max_0_0_11(P,__PREV) :- __chain__max_0_0_11(P,__NEXT); __next_0__dom___max_0_0_11(__PREV,__NEXT).
        weight = elem.terms[0]

        ret = list(self.domain_predicates.create_domain(new_predicate))
        anon = AnnotatedPredicate(
            self.domain_predicates.domain_predicate(new_predicate), tuple(range(0, new_predicate.arity))
        )
        ret.extend(self.domain_predicates.create_next_pred_for_annotated_pred(anon, 0))
        # create my own chaining rules

        minmax_pred = self.domain_predicates.max_anon_predicate(anon, 0)
        if agg.atom.function == AggregateFunction.Max:
            minmax_pred = self.domain_predicates.min_anon_predicate(anon, 0)

        chain_name = self.domain_predicates.chain_pred(anon, 0, agg.atom.function == AggregateFunction.Max).name
        next_pred = self.domain_predicates.next_anon_predicate(anon, 0)

        aux_head = Literal(
            LOC,
            Sign.NoSign,
            SymbolicAtom(Function(LOC, chain_name, rest_vars + [weight], False)),
        )
        ret.append(Rule(LOC, aux_head, list(chain(elem.condition, lits_with_vars))))

        prev_agg = NEXT
        next_agg = PREV
        if agg.atom.function == AggregateFunction.Max:
            prev_agg = PREV
            next_agg = NEXT

        head = Literal(
            LOC,
            Sign.NoSign,
            SymbolicAtom(Function(LOC, chain_name, rest_vars + [prev_agg], False)),
        )

        body = []
        body.append(
            Literal(
                LOC,
                Sign.NoSign,
                SymbolicAtom(Function(LOC, chain_name, rest_vars + [next_agg], False)),
            )
        )
        body.append(
            Literal(
                LOC,
                Sign.NoSign,
                SymbolicAtom(Function(LOC, next_pred.name, [PREV, NEXT], False)),
            )
        )
        ret.append(Rule(LOC, head, body))

        # 3. create actual new max/min predicate
        # __max_0_0_11(P,__PREV) :- __chain__max_0_0_11(P,__PREV);
        #                             not __chain__max_0_0_11(P,__NEXT): __next_0__dom___max_0_0_11(__PREV,__NEXT).
        # __max_0_0_11(P,#inf) :- __min_0__dom___max_0_0_11(X); not __chain__max_0_0_11(P,X); person(P).
        head = Literal(
            LOC,
            Sign.NoSign,
            SymbolicAtom(Function(LOC, new_predicate.name, rest_vars + [prev_agg], False)),
        )

        body = []
        body.append(
            Literal(
                LOC,
                Sign.NoSign,
                SymbolicAtom(Function(LOC, chain_name, rest_vars + [prev_agg], False)),
            )
        )
        body.append(
            ConditionalLiteral(
                LOC,
                Literal(
                    LOC,
                    Sign.Negation,
                    SymbolicAtom(Function(LOC, chain_name, rest_vars + [next_agg], False)),
                ),
                [
                    Literal(
                        LOC,
                        Sign.NoSign,
                        SymbolicAtom(Function(LOC, next_pred.name, [PREV, NEXT], False)),
                    )
                ],
            )
        )
        ret.append(Rule(LOC, head, body))

        border = Supremum
        if agg.atom.function == AggregateFunction.Max:
            border = Infimum

        head = Literal(
            LOC,
            Sign.NoSign,
            SymbolicAtom(Function(LOC, new_predicate.name, rest_vars + [SymbolicTerm(LOC, border)], False)),
        )

        body = []
        var_x = Variable(LOC, "X")

        body.append(
            Literal(
                LOC,
                Sign.NoSign,
                SymbolicAtom(Function(LOC, minmax_pred.name, [var_x], False)),
            )
        )
        body.append(
            Literal(
                LOC,
                Sign.Negation,
                SymbolicAtom(Function(LOC, chain_name, rest_vars + [var_x], False)),
            )
        )
        body.extend(lits_with_vars)
        ret.append(Rule(LOC, head, body))
        return ret

    def _simple_translation(self, rule: AST, agg: AST) -> list[AST]:
        """translate a min/max aggregate that has a <,<=,>,>= comparison in the same
        direction to a simple set of rules"""
        ret = []
        gvars = global_vars_inside_body(rule.body)
        uv = UniqueVariables(rule)
        body = rule.body
        body.remove(agg)
        for elem in agg.atom.elements:
            lvars = set(collect_ast(elem, "Variable")) - gvars
            old2new = {oldvar: uv.make_unique(oldvar) for oldvar in lvars}

            elem = transform_ast(elem, "Variable", lambda x, old2new=old2new: old2new.setdefault(x, x))  # type: ignore
            newlits = elem.condition
            newlits.append(
                Literal(
                    LOC,
                    agg.sign,
                    Comparison(agg.atom.left_guard.term, [Guard(agg.atom.left_guard.comparison, elem.terms[0])]),
                )
            )
            rule = rule.update(body=list(body) + list(newlits))
            ret.append(rule)
        return ret

    def replace_orig(
        self, rule: AST, agg: AST, new_name: str, rest_vars: list[AST], lits_without_vars: list[AST]
    ) -> list[AST]:
        """replace the original aggregate inside the rule with the new named predicate"""
        body: list[AST] = []
        ret: list[AST] = []

        analytics = AggAnalytics(agg.atom)
        max_var = Variable(LOC, f"__VAR{new_name}")
        if analytics.equal_variable_bound:
            max_var = Variable(LOC, analytics.equal_variable_bound.pop(0))
        body.append(
            Literal(
                LOC,
                Sign.NoSign,
                SymbolicAtom(Function(LOC, new_name, rest_vars + [max_var], False)),
            )
        )
        # repair bounds
        if len(analytics.equal_variable_bound) == 1:
            body.append(
                Literal(
                    LOC,
                    Sign.NoSign,
                    Comparison(
                        max_var,
                        [
                            Guard(
                                ComparisonOperator.Equal,
                                Variable(LOC, analytics.equal_variable_bound[0]),
                            )
                        ],
                    ),
                )
            )
        for bound in analytics.bounds:
            body.append(Literal(LOC, Sign.NoSign, Comparison(max_var, [bound])))
        body.extend(lits_without_vars)
        ret.append(rule.update(body=body))
        if rule.ast_type == ASTType.Rule:
            self._store_aggregate_head(agg.atom.function, rule.head, rest_vars, max_var, new_name)
        else:
            self._store_aggregate_for_minimize(agg.atom.function, rest_vars, max_var, new_name)
        return ret

    def _chain_translation(self, rule: AST, agg: AST) -> list[AST]:
        """translate a single elemented min/max aggregate to chaining rules"""
        if len(agg.atom.elements) > 1:
            log.info(
                f"Cannot translate {loc2str(agg.location)} as multiple elements "
                "inside min/max aggregate are not yet supported. See #9."
            )
            return [rule]
        number_of_aggregate = 0
        assert len(agg.atom.elements) == 1
        elem = agg.atom.elements[0]
        weight = elem.terms[0]
        direction = "min"
        if agg.atom.function == AggregateFunction.Max:
            direction = "max"

        # 1. create a new domain for the complete elem.condition + lits_with_vars
        new_name = f"__{direction}_{number_of_aggregate}_{str(rule.location.begin.line)}"
        new_predicate = Predicate(new_name, 1)

        head = SymbolicAtom(Function(LOC, new_name, [weight], False))
        lits_with_vars = []
        lits_without_vars = []
        rest_vars: set[AST] = set()  # variable names that are used in- but also outside of the aggregate
        inside_variables = set(chain(*map(lambda x: collect_ast(x, "Variable"), agg.atom.elements)))
        for blit in rule.body:
            if blit == agg:
                continue
            blit_vars = set(collect_ast(blit, "Variable"))
            if len(blit_vars.intersection(inside_variables)) != 0:
                rest_vars.update(blit_vars)
                lits_with_vars.append(blit)
            else:
                lits_without_vars.append(blit)
        if rule.ast_type == ASTType.Minimize:
            rest_vars.update(inside_variables.intersection(collect_ast(rule.weight, "Variable")))
            rest_vars.update(inside_variables.intersection(collect_ast(rule.priority, "Variable")))
            for t in rule.terms:
                rest_vars.update(inside_variables.intersection(collect_ast(t, "Variable")))
        # variables that are used inside but also outside of the aggregate
        rest_vars_sorted: list[AST] = sorted(rest_vars)

        self.domain_predicates.add_domain_rule(
            Predicate(new_name, 1), [(head, list(chain(elem.condition, lits_with_vars)))]
        )
        if not self.domain_predicates.has_domain(new_predicate):
            log.info(
                f"Cannot translate {loc2str(agg.location)} as I cannot infer "
                f"a domain for {[str(x) for x in list(chain(elem.condition, lits_with_vars))]}."
            )
            return [rule]

        ret = self._create_aggregate_replacement(agg, elem, rest_vars_sorted, new_predicate, lits_with_vars)

        # 3. replace original rule or minimize
        return ret + self.replace_orig(rule, agg, new_name, rest_vars_sorted, lits_without_vars)

    def _process_rule(self, rule: AST) -> list[AST]:
        """returns a list of rules to replace this rule"""

        # TODO: currently only one aggregate is translated, create loop
        #  lift this restriction, could have more than one agg,
        # but the number needs to go into the chain name
        agg = self._minmax_agg(rule)
        if agg is None:
            return [rule]

        if not any(map(self._translatable_element, agg.atom.elements)):
            return [rule]  # nocoverage, #issue9

        if not agg.atom.right_guard:
            lt = agg.atom.left_guard.comparison in (ComparisonOperator.LessThan, ComparisonOperator.LessEqual)
            gt = agg.atom.left_guard.comparison in (ComparisonOperator.GreaterThan, ComparisonOperator.GreaterEqual)
            if (  # pylint: disable=too-many-boolean-expressions
                agg.sign == Sign.NoSign
                and (
                    (agg.atom.function == AggregateFunction.Max and lt)
                    or (agg.atom.function == AggregateFunction.Min and gt)
                )
            ) or (
                agg.sign == Sign.Negation
                and (
                    (agg.atom.function == AggregateFunction.Min and lt)
                    or (agg.atom.function == AggregateFunction.Max and gt)
                )
            ):
                return self._simple_translation(rule, agg)
        return self._chain_translation(rule, agg)

    def _create_replacement(
        self,
        minmaxpred: MinMaxPred,
        minimize: bool,
        terms: list[AST],
        oldmax: AST,
        rest_cond: list[AST],
        function: Callable[[AST, list[AST], list[AST]], AST],
    ) -> list[AST]:
        assert minmaxpred is not None

        if minimize:

            def negate_if(x: AST) -> AST:
                return x

        else:

            def negate_if(x: AST) -> AST:
                return UnaryOperation(LOC, UnaryOperator.Minus, x)

        aggtype, translation, idx = minmaxpred

        if aggtype == AggregateFunction.Max:
            prev = PREV
            next_ = NEXT
        else:
            prev = NEXT
            next_ = PREV

        # (__NEXT-__PREV), __chain__max_0_0_x(__PREV,__NEXT) : __chain__max_0_0_x(P,__NEXT),
        #  __next_0__dom___max_0_0_11(__PREV,__NEXT)
        weight = negate_if(BinaryOperation(LOC, BinaryOperator.Minus, next_, prev))
        newpred = translation.newpred
        chain_name = self.domain_predicates.chain_pred(
            AnnotatedPredicate(Predicate(newpred.name, 1), (0,)), 0, aggtype == AggregateFunction.Max
        ).name
        new_terms = [Function(LOC, chain_name, [PREV, NEXT], False)] + list(terms)

        newargs = translation.translate_parameters(oldmax.atom.symbol.arguments)
        newargs = [next_ if i == idx else x for i, x in enumerate(newargs)]
        for arg in newargs:
            assert isinstance(arg, AST)
        chainpred = Literal(
            LOC,
            Sign.NoSign,
            SymbolicAtom(Function(LOC, chain_name, newargs, False)),  # type: ignore
        )
        dompred = self.domain_predicates.dom_named_predicate(newpred.name, 1)
        anon = AnnotatedPredicate(dompred, tuple(range(0, dompred.arity)))
        nextpred = Literal(
            LOC,
            Sign.NoSign,
            SymbolicAtom(
                Function(
                    LOC,
                    self.domain_predicates.next_anon_predicate(anon, 0).name,
                    [PREV, NEXT],
                    False,
                )
            ),
        )
        ret = []
        ret.append(function(weight, new_terms, [chainpred, nextpred] + rest_cond))
        # __NEXT, __chain__max_0_0_x(#sup,__NEXT) : __chain__max_0_0_x(P,__NEXT), __min_0__dom___max_0_0_x(__NEXT)
        infsup = Infimum
        if aggtype == AggregateFunction.Max:
            infsup = Supremum
        weight = negate_if(next_)
        new_terms = [Function(LOC, chain_name, [SymbolicTerm(LOC, infsup), next_], False)] + list(terms)
        if aggtype == AggregateFunction.Max:
            name = self.domain_predicates.min_anon_predicate(anon, 0).name
        else:
            name = self.domain_predicates.max_anon_predicate(anon, 0).name
        minmaxlit = Literal(
            LOC,
            Sign.NoSign,
            SymbolicAtom(
                Function(
                    LOC,
                    name,
                    [next_],
                    False,
                )
            ),
        )
        ret.append(function(weight, new_terms, [chainpred, minmaxlit] + rest_cond))
        # #inf and #sup are ignored by minimized and therefore not included
        # (also would require more complex variable bindings)
        return ret

    def _replace_results_in_minimize(self, stm: AST, minimizes: dict[tuple[AST, ...], list[AST]]) -> list[AST]:
        """
        return list of statements that replaces the minimize statement
        or returns the statement itself in a list
        """
        # pylint: disable=too-many-branches
        assert stm.ast_type == ASTType.Minimize
        if not self._minmax_preds:
            return [stm]

        term_tuple = (
            stm.weight,
            stm.priority,
            *stm.terms,
        )
        if stm.weight.ast_type == ASTType.Variable:
            varname = stm.weight.name
            minimize = True
        elif (
            stm.weight.ast_type == ASTType.UnaryOperation
            and stm.weight.operator_type == UnaryOperator.Minus
            and stm.weight.argument.ast_type == ASTType.Variable
        ):
            varname = stm.weight.argument.name
            minimize = False
        else:
            log.info(f"Cannot optimize {loc2str(stm.location)} as the weight is not simple enough.")
            return [stm]

        preds = [
            x.pred
            for x in set(chain.from_iterable(predicates(b, {Sign.NoSign, Sign.DoubleNegation}) for b in stm.body))
        ]
        minmaxpred: Optional[MinMaxAggregator.MinMaxPred] = None
        for aggtype, translation, idx in self._minmax_preds:
            if translation.oldpred in preds:
                # check if it is globally safe to assume a unique tuple semantics

                unsafe = []
                for terms, objective in minimizes.items():
                    if potentially_unifying_sequence(terms, term_tuple):
                        unsafe.extend([x for x in objective if x != stm])

                if not unsafe:
                    minmaxpred = (aggtype, translation, idx)
                    break
        if minmaxpred is None:
            log.info(f"Cannot optimize {loc2str(stm.location)} as the tuple is not globally unique.")
            return [stm]

        rest_cond: list[AST] = []
        oldmax: Optional[AST] = None
        for cond in stm.body:
            if minmaxpred is not None and list(map(lambda x: x.pred, predicates(cond, {Sign.NoSign}))) == [
                minmaxpred[1].oldpred
            ]:
                oldmax = cond
            else:
                rest_cond.append(cond)
        assert oldmax is not None

        # check if all Variables from old predicate are used in the tuple identifier
        # to make a unique semantics
        # see issue #8
        old_vars = set(map(lambda x: x.name, collect_ast(oldmax, "Variable"))) - {varname}
        term_vars = {x.name for x in chain.from_iterable(map(_characteristic_variables, term_tuple[2:]))}
        if not old_vars <= term_vars:
            log.info(
                f"Cannot use chaining in {loc2str(stm.location)} as not all variables {old_vars} are used in the tuple."
                " (#8)"
            )
            return [stm]

        replacement = self._create_replacement(
            minmaxpred,
            minimize,
            stm.terms,
            oldmax,
            rest_cond,
            lambda weight, terms, conditions: Minimize(LOC, weight, stm.priority, terms, conditions),
        )
        return replacement

    def _split_element(
        self, loc: Location, elem: AST, rest_elems: list[AST]
    ) -> tuple[Optional[AST], Optional[MinMaxPred], list[AST]]:
        """
        splits element conditions into the first valid min/max predicate it finds and the other conditions
        returns it as triple
        (max predicate, translation, rest of the conditions)
        """
        preds = [
            x.pred
            for x in set(chain.from_iterable(predicates(b, {Sign.NoSign, Sign.DoubleNegation}) for b in elem.condition))
        ]
        rest_cond: list[AST] = []
        minmaxpred: Optional[MinMaxAggregator.MinMaxPred] = None
        oldmax: Optional[AST] = None
        for aggtype, translation, idx in self._minmax_preds:
            if translation.oldpred in preds:
                # check if it is locally safe to assume a unique tuple semantics

                # check if any of the other elements is potentially unifying and therefore unsafe
                # pylint: disable=cell-var-from-loop
                if not any(map(lambda x: potentially_unifying_sequence(x.terms, elem.terms), rest_elems)):
                    minmaxpred = (aggtype, translation, idx)
                    break
                log.info(f"Cannot optimize {loc2str(loc)} as the tuple is not unique.")

        for cond in elem.condition:
            if minmaxpred is not None and list(map(lambda x: x.pred, predicates(cond, {Sign.NoSign}))) == [
                minmaxpred[1].oldpred
            ]:
                oldmax = cond
            else:
                rest_cond.append(cond)
        return oldmax, minmaxpred, rest_cond

    def _replace_results_in_sum_agg_elem(self, elem: AST, rest_elems: list[AST]) -> list[AST]:
        """
        replaces min/max predicates in sum aggregate elements by returning a
        new list of elements
        """
        term_tuple = elem.terms
        # split condition into the max predicate + translation and the rest
        old_max, minmaxpred, rest_cond = self._split_element(term_tuple[0].location, elem, rest_elems)
        if minmaxpred is None:
            return [elem]
        assert old_max is not None

        if term_tuple[0].ast_type == ASTType.Variable:
            varname = term_tuple[0].name
            minimize = True  # TODO: find a better name
        elif (
            term_tuple[0].ast_type == ASTType.UnaryOperation
            and term_tuple[0].operator_type == UnaryOperator.Minus
            and term_tuple[0].argument.ast_type == ASTType.Variable
        ):
            varname = term_tuple[0].argument.name
            minimize = False

        else:
            log.info(f"Cannot optimize {loc2str(term_tuple[0].location)} as the weight is not simple enough.")
            return [elem]

        # check if all Variables from old predicate are used in the tuple identifier
        # to make a unique semantics
        # NOTE: is this check really useful ? Why
        old_vars = set(map(lambda x: x.name, collect_ast(old_max, "Variable"))) - {varname}
        term_vars = {x.name for x in chain.from_iterable(map(_characteristic_variables, term_tuple[1:]))}
        if not old_vars <= term_vars:
            log.info(
                f"Cannot optimize {loc2str(term_tuple[0].location)} as not all variables {old_vars}"
                " are used in the tuple."
                "This might be a bug, see issue #8."
            )
            return [elem]

        replacement = self._create_replacement(
            minmaxpred,
            minimize,
            term_tuple[1:],
            old_max,
            rest_cond,
            lambda weight, terms, conditions: BodyAggregateElement([weight] + terms, conditions),
        )
        return replacement

    def _replace_results_in_sum_agg(self, agg: AST) -> AST:
        """
        replaces min/max predicates in sum aggregates by returning an aggregate
        (this might the a new one or the old one)
        """
        atom = agg.atom
        elements = []
        for elem in atom.elements:
            elements.extend(self._replace_results_in_sum_agg_elem(elem, [x for x in atom.elements if x != elem]))
        return Literal(LOC, agg.sign, BodyAggregate(LOC, atom.left_guard, atom.function, elements, atom.right_guard))

    # NOTE. this might actually not be advantageous as it produces a larger set of potential sums
    def _replace_results_in_sum(self, stm: AST) -> list[AST]:
        """
        replaces min/max predicates in sum aggregates by returning a list of statements
        or the old statement if no replacement is possible
        """

        assert stm.ast_type == ASTType.Rule
        body = []
        for b in stm.body:
            if (
                b.ast_type == ASTType.Literal
                and b.atom.ast_type == ASTType.BodyAggregate
                and b.atom.function in (AggregateFunction.Sum, AggregateFunction.SumPlus)
            ):
                body.append(self._replace_results_in_sum_agg(b))
            else:
                body.append(b)
        return [Rule(LOC, stm.head, body)]

    def _replace_results_in_x(self, prg: list[AST], minimizes: dict[tuple[AST, ...], list[AST]]) -> list[AST]:
        """
        replace all predicates that computed max/minimum values with their order encoding
        in sum contexts
        """
        ret = []
        for stm in prg:
            if stm.ast_type == ASTType.Minimize:
                ret.extend(self._replace_results_in_minimize(stm, minimizes))
            elif stm.ast_type == ASTType.Rule and any(
                map(
                    lambda body: body.ast_type == ASTType.Literal
                    and body.atom.ast_type == ASTType.BodyAggregate
                    and body.atom.function in (AggregateFunction.Sum, AggregateFunction.SumPlus),
                    stm.body,
                )
            ):
                ret.extend(self._replace_results_in_sum(stm))
            else:
                ret.append(stm)
                continue
        return ret

    def execute(self, prg: list[AST]) -> list[AST]:
        """
        replace easy min/max aggregates with chaining rules
        also replace the usage of the results in sum and optimize conditions
        """
        ret: list[AST] = []
        minimizes: dict[tuple[AST, ...], list[AST]] = defaultdict(list)
        for rule in prg:
            if rule.ast_type not in (ASTType.Rule, ASTType.Minimize):
                ret.append(rule)
                continue

            for new_rule in self._process_rule(rule):
                if new_rule.ast_type == ASTType.Minimize:
                    minimizes[
                        (
                            new_rule.weight,
                            new_rule.priority,
                            *new_rule.terms,
                        )
                    ].append(new_rule)
                ret.append(new_rule)

        ret = self._replace_results_in_x(ret, minimizes)
        return ret



""" test ast utility functions """
import pytest
from clingo.ast import AST, Variable, parse_string

from ngo.normalize import preprocess
from ngo.utils.ast import LOC, Predicate
from ngo.utils.globals import UniqueNames, UniqueVariables, auto_detect_input, auto_detect_output


def test_unique_names() -> None:
    """check unique name creation"""
    prg = """
a.
f(x,y) :- __aux_1(a,b,c).
    """

    ast: list[AST] = []
    parse_string(prg, ast.append)
    ast = preprocess(ast)
    unique_names = UniqueNames(ast, [])
    assert unique_names.new_predicate("b", 0) == Predicate("b", 0)
    assert unique_names.new_predicate("a", 0) == Predicate("a1", 0)
    assert unique_names.new_auxpredicate(3) == Predicate("__aux_2", 3)


def test_unique_variables() -> None:
    """check unique name creation"""
    prg = """
f(X,C) :- b(X,A,B).
    """

    ast: list[AST] = []
    parse_string(prg, ast.append)
    ast = preprocess(ast)
    unique_names = UniqueVariables(ast[1])
    assert unique_names.make_unique(Variable(LOC, "X")).name == "X0"
    assert unique_names.make_unique(Variable(LOC, "Y")).name == "Y"
    assert unique_names.make_unique(Variable(LOC, "Y")).name == "Y0"
    assert unique_names.make_unique(Variable(LOC, "Y")).name == "Y1"
    assert unique_names.make_unique(Variable(LOC, "A")).name == "A0"


@pytest.mark.parametrize(
    "lhs, input_predicates",
    (
        (
            """
a.
{b(X) : c(X)}.
:- d, not b(X).
:- not e, e(X).
#minimize {1,2,3 : f(X)}.
wall(X,Y) :- wall(Y,X).
            """,
            {
                Predicate("f", 1),
                Predicate("c", 1),
                Predicate("d", 0),
                Predicate("e", 0),
                Predicate("e", 1),
                Predicate("wall", 2),
            },
        ),
    ),
)
def test_auto_input(lhs: str, input_predicates: set[Predicate]) -> None:
    """test removal of superseeded literals on whole programs"""
    ast: list[AST] = []
    parse_string(lhs, ast.append)
    ast = preprocess(ast)
    auto_detected = auto_detect_input(ast)
    assert input_predicates == set(auto_detected)


@pytest.mark.parametrize(
    "lhs, output_predicates",
    (
        (
            """
#show p/3.
#show a : b.
#show c(X) : d(X).
            """,
            {Predicate("p", 3), Predicate("b", 0), Predicate("d", 1)},
        ),
    ),
)
def test_auto_output(lhs: str, output_predicates: set[Predicate]) -> None:
    """test removal of superseeded literals on whole programs"""
    ast: list[AST] = []
    parse_string(lhs, ast.append)
    ast = preprocess(ast)
    auto_detected = auto_detect_output(ast)
    assert output_predicates == set(auto_detected)



"""
The main entry point for the application.
"""
import logging
import sys
from functools import partial
from typing import Callable

from clingo import MessageCode
from clingo.ast import AST, parse_files

from ngo.api import optimize
from ngo.utils.globals import auto_detect_input, auto_detect_output
from ngo.utils.parser import get_parser


def _wrap_logger(logger: Callable[[str], None], _: MessageCode, message: str) -> None:
    logger(message)


def main() -> None:
    """
    Run the main function.
    """
    # pylint: disable=too-many-branches # will be refactored
    parser = get_parser()
    args = parser.parse_args()
    logging.basicConfig(stream=sys.stderr, level=args.log)

    prg: list[AST] = []
    parse_files(["-"], prg.append, logger=partial(_wrap_logger, logging.warning))
    ### create general tooling and analyzing classes
    if args.input_predicates == "auto":
        args.input_predicates = auto_detect_input(prg)
    if args.output_predicates == "auto":
        args.output_predicates = auto_detect_output(prg)
    elif args.output_predicates == "":
        args.output_predicates = []

    prg = optimize(
        prg,
        args.input_predicates,
        args.output_predicates,
        cleanup="cleanup" in args.enable,
        unused="unused" in args.enable,
        duplication="duplication" in args.enable,
        symmetry="symmetry" in args.enable,
        minmax_chains="minmax_chains" in args.enable,
        sum_chains="sum_chains" in args.enable,
        math="math" in args.enable,
        inline="inline" in args.enable,
        projection="projection" in args.enable,
    )

    for i in prg:
        print(i)


if __name__ == "__main__":
    main()



# pylint: disable=line-too-long
"""
Impress your friends with faster encodings.
ngo is a program and python library to optimize non ground logic ASP encodings for performance.

# Install

You can install the current development version using pip

```shell
pip install ngo
```

# Quickstart - Command line version
ngo converts an encoding read from stdin to an optimized encoding.

```shell
cat encoding.lp | ngo > new_encoding.lp
```

To get a better view of what ngo did to your encoding compare the two files
```shell
cat encoding.lp | ngo --enable none > orig_encoding.lp
cat encoding.lp | ngo > optimized_encoding.lp
```
Now use any diff viewer on the two new files.

### Input Predicates
For some traits you have to state which predicates in your logic
program are considered input. This means they are provided outside of the encoding,
for example in an instance file. You give a comma separated list of predicates of the form `name/n` where `n` is the arity of the predicate. For constants you can use arity 0. 
Example:
```shell
ngo --input-predicates="edge/2, node/1"
```
You can also write `--input-predicates=auto` (its the default) to try to auto-detect the input predicates by screening your program for all predicates and taking the ones that do not occur in the head of any rule.


### Output Predicates
For some traits you have to state which predicates in your logic
program are considered output. This means they are not allowed to be removed or changed by ngo.
You give a comma separated list of predicates of the form `name/n` where `n` is the arity of the predicate. For constants you can use arity 0. 
Example:
```shell
ngo --output-predicates="path/2, cost/1"
```
You can also write `--output-predicates=auto` (its the default). It will detect the predicates from the show statements. If performance is critical it makes sense to remove all kinds of output rewriting from the encoding and make a dedicated postprocessing step for it. For this consider using `--output-predicates` without an argument to make it empty.

### Option --enable

The option enable support several [traits](#traits) that can be added, like
```shell
cat encoding.lp | ngo --enable math minmax_chains
```
By default this setting is set to `default`, which enables most traits.
See [here](#traits) for a detailed information of what each trait does.
To just rewrite your program into std clingo.ast normal form, use `none` to disable any optimizations.
You can use `all` to enable all available traits.

# Quickstart - API version

You can use the `optimize` function to optimize the list of AST's constituting your encoding.
ngo needs just one additional line of code.

See
```python
from clingo import Control
from clingo.ast import parse_files, ProgramBuilder
from ngo import optimize, auto_detect_input, auto_detect_output

ctl = Control()

prg = []
parse_files(["encoding.lp"], prg.append)
# this line optimizes your encoding
prg = optimize(prg, auto_detect_input(prg), auto_detect_output(prg))

with ProgramBuilder(ctl) as bld:
    for stm in prg:
        # print(stm) # shows the optimized rules
        bld.add(stm)

ctl.load("instance.lp")
ctl.ground([('base', [])])
print(ctl.solve(on_model=print))
```


# Traits

The following traits are available:

## cleanup

This module removes unnecessary literals that are superseeded by others.
This requires the knowledge of which predicates are input predicates,
beeing derived somewhere else, e.g. an instance of the problem.

```Prolog
b(X,Y) :- dom(X), dom(Y), X+Y < 42.
a(X,Y) :- b(X,Y), dom(X), dom(Y).
```
becomes
```Prolog
b(X,Y) :- dom(X), dom(Y), X+Y < 42.
a(X,Y) :- b(X,Y).
```
if `dom/1` is an [input predicates](#input-predicates).

## unused

This module removes unnecessary variables from predicates. It might shrink predicates down and reduce their arity.
If a predicate is not used in a constraint not in the `--output-predicates` it might be removed completely.
So, this program:
```Prolog
b(X) :- c(X).
{ a } :- b(X).
foo :- a.
```
where `c/1` is an [input predicates](#input-predicates), becomes:
```Prolog
{ a } :- c(_).
```

## duplication

This module replaces sets of literals that occur in multiple rules
 with an extra predicate that is derived only once.
```Prolog
foo :- a, b, c.
bar :- a, b, d.
foobar :- {e : a, b}.
```
becomes
```Prolog
__aux_1 :- a; b.
foo :- c; __aux_1.
bar :- d; __aux_1.
foobar :- { e: __aux_1 }.
```

## symmetry

This trait replaces bodys with symmetry breaking rules that
count different occurences of the same predicate with a counting aggregate
if this preserves semantics. This can reduce the number of grounded
rules.

```Prolog
:- slot(J1,M,T); slot(J2,M,T); J1 != J2.
```
gets replaced by
```Prolog
:- slot(_,M,T), 2 #count {J1 : slot(J1,M,T)}.
```
For more complex rules the symmetry breaking is improved:
```Prolog
:- slot(J1,M,T1); slot(J2,M,T2); J1 != J2, T1 != T2.
```
becomes
```Prolog
:- slot(J1,M,T1); slot(J2,M,T2); J1 < J2, T1 != T2.
```

## minmax_chains

Replaces min/max aggregates with a chain encoding.
Replaces the result of min/max aggregates with chains in optimize statements and in sum aggregates.
Also replaces simple bounded min/max aggregates with simple rules.

## sum_chains

Detects partial predicates that under an at_most one restriction.
Replaces these predicates in sum aggregates/optimization statements with a chain encoding.

```Prolog
{ shift(D,L) : pshift(D,L) } 1 :- day(D).
a(X) :- X = #sum {L,D : shift(D,L)}.
```
In this example only one shift length per day can exists.
Therefore the shift length is computed using chaining and then the
parts of the chain are used inside the #sum aggregate.
This can reduce grounding and solving time.
Depending on the domain of the weight Variable (L in this case here)
the computation of the ordered domain can cause problems in grounding.

## math

Tries to optimize any form of comparisons and aggregates in the rule body.
This means that:
```Prolog
a :- X = #sum { 1,a : a }, Y=#sum{ 1,b: b }, X+Y=2.
```
gets replaces by something similar to
```
a :- 0 = #sum { 1,a: a; 1,b: b; -2 }.
```Prolog
This can reduce grounding drastically and might have an effect on solving.

## inline

This trait inlines rules that use aggregates into other aggregates or objective statements.

```Prolog
suminline(A,B) :- a(A); B = #sum { Y: person(A,Y) }.
foo(X) :- X = #sum { F,V: suminline(V,F); A: test(A,B) }.
```
becomes
```Prolog
foo(X) :- X = #sum { A: test(A,B); Y,V: a(V), person(V,Y) }.
```
This is only possible if it is safe to change set semantics using the tuples in the aggregates.
This trait also works for minimize statements and weak constraints.

## projection

This trait splits rules in a way that the grounding complexity is reduced.
```Prolog
p(A,D) :- q(A,B,C), r(A D), t(E), not s(B,E).
```
becomes
```Prolog
aux(A) :- q(A,B,C), t(E), not s(B,E).
p(A,D) :- aux(A), r(A D).
```

This can reduce grounding drastically but introduces additional solving variables.
"""


from ngo.api import optimize
from ngo.utils.ast import Predicate
from ngo.utils.globals import auto_detect_input, auto_detect_output

__all__ = ["optimize", "auto_detect_input", "auto_detect_output", "Predicate"]



"""
 This module removes literals that are subsumed by others
"""

import logging
from collections import defaultdict
from dataclasses import dataclass
from itertools import permutations
from typing import Iterable, Iterator, Optional

from clingo.ast import AST, ASTType, Sign

from ngo.normalize import inline_arithmetic
from ngo.utils.ast import Predicate, SignedPredicate, headderivable_predicates, is_predicate

log = logging.getLogger(__name__)


@dataclass(frozen=True, eq=True, unsafe_hash=True)
class Mapping:
    """store a mapping from the head predicate to the superseeded body predicate
    the var_map maps positions to each other,
    eg var_map(1,1,3)
    means head_pred(X,f(Y),Z*Y) and body(X,X,Y*Z)
    """

    head_pred: Predicate
    body_pred: SignedPredicate
    var_map: tuple[int, ...]


class CleanupTranslator:
    """Removes superseeded literals in the body, conditions"""

    def __init__(self, input_predicates: list[Predicate]):
        self.input_predicates = input_predicates
        self.superseeds: set[Mapping] = set()

    @staticmethod
    def _create_mappings(head_symbol: AST, body_lits: list[AST]) -> Iterator[Mapping]:
        head_pred = Predicate(head_symbol.name, len(head_symbol.arguments))
        for cond in body_lits:
            assert cond.ast_type == ASTType.Literal
            if is_predicate(cond):
                body_symbol = cond.atom.symbol
                var_map: list[int] = []
                for arg in body_symbol.arguments:
                    if arg in head_symbol.arguments:
                        var_map.append(head_symbol.arguments.index(arg))
                if len(var_map) == len(body_symbol.arguments):
                    yield Mapping(
                        head_pred,
                        SignedPredicate(cond.sign, Predicate(body_symbol.name, len(body_symbol.arguments))),
                        tuple(var_map),
                    )

    @staticmethod
    def _collect_top_level_body_symbols(body: Iterator[AST]) -> Iterator[AST]:
        for lit in body:
            if (
                lit.ast_type == ASTType.Literal
                and lit.atom.ast_type == ASTType.SymbolicAtom
                and lit.atom.symbol.ast_type == ASTType.Function
            ):
                yield lit

    def _compute_local_superseed(self, pred: Predicate, rule: AST) -> set[Mapping]:
        local_superseed: set[Mapping] = set()
        head_symbols: list[AST] = []
        assert rule.ast_type == ASTType.Rule
        head = rule.head
        if is_predicate(head):
            symbol = head.atom.symbol
            if pred == Predicate(symbol.name, len(symbol.arguments)):
                head_symbols.append(symbol)
        elif head.ast_type == ASTType.HeadAggregate:
            for element in head.elements:
                assert element.condition.ast_type == ASTType.ConditionalLiteral
                lit = element.condition.literal
                assert lit.ast_type == ASTType.Literal
                if lit.atom.ast_type == ASTType.SymbolicAtom and lit.atom.symbol.ast_type == ASTType.Function:
                    symbol = lit.atom.symbol
                    if pred != Predicate(symbol.name, len(symbol.arguments)):
                        continue
                    head_symbols.append(symbol)
                    local_superseed.update(set(x for x in self._create_mappings(symbol, element.condition.condition)))
        elif head.ast_type in (ASTType.Aggregate, ASTType.Disjunction):
            for element in head.elements:
                lit = element.literal
                assert lit.ast_type == ASTType.Literal
                if is_predicate(lit):
                    symbol = lit.atom.symbol
                    if pred != Predicate(symbol.name, len(symbol.arguments)):
                        continue
                    head_symbols.append(symbol)
                    local_superseed.update(set(x for x in self._create_mappings(symbol, element.condition)))
        # add all body elements to all heads according to they variables
        body_literals = list(self._collect_top_level_body_symbols(rule.body))
        for symbol in head_symbols:
            local_superseed.update(set(x for x in self._create_mappings(symbol, body_literals)))
        return local_superseed

    @staticmethod
    def transitive_closure(a: set[Mapping]) -> set[Mapping]:
        """
        Mappings are transitive, so this generates the closure explicitly
        a(X,Y,Z) : b(Z,Y,X)  [3,2,1]
        b(A,B,C) : c(C,A)  [3,1]

        a(X,Y,Z) : b(Z,Y,X) : c(X,Z) [1,3]
        """
        closure = set(a)
        while True:
            new_relations = set()
            for lhs in closure:
                for rhs in closure:
                    if lhs.body_pred.sign == Sign.NoSign and lhs.body_pred.pred == rhs.head_pred:
                        var_map = [lhs.var_map[m] for m in rhs.var_map]
                        new_relations.add(Mapping(lhs.head_pred, rhs.body_pred, tuple(var_map)))
            closure_until_now = closure | new_relations
            if closure_until_now == closure:
                break
            closure = closure_until_now
        return closure

    def _find_superseeded(self, prg: list[AST]) -> None:
        """fill self.superseeded with predicates that superseed others"""
        # 1. create a mapping for head derivable predicates and their rules
        pred2rules: dict[Predicate, list[int]] = defaultdict(list)
        for index, stm in enumerate(prg):
            for spred in headderivable_predicates(stm):
                if spred.pred not in self.input_predicates:
                    pred2rules[spred.pred].append(index)
        for pred, rule_ids in pred2rules.items():
            # 2. given a predicate and a set of rules
            # for all rules, find all mappings that superseed locally
            # the intersection of local sets is the superseed set

            superseed: Optional[set[Mapping]] = None
            for id_ in rule_ids:
                rule = prg[id_]
                if superseed is None:
                    superseed = self._compute_local_superseed(pred, rule)
                else:
                    superseed.intersection_update(self._compute_local_superseed(pred, rule))
            assert isinstance(superseed, set)
            self.superseeds.update(superseed)
        # 3. superseeded is transitive, make it explicit
        self.superseeds = self.transitive_closure(self.superseeds)

    def _superseeded(self, lhs: AST, rhs: AST) -> bool:
        """use the mappings to check if lhs Literal superseeds rhs literal"""
        if not (is_predicate(lhs) and lhs.sign == Sign.NoSign):
            return False
        if not is_predicate(rhs):
            return False
        lhs_symbol = lhs.atom.symbol
        lhs_pred = Predicate(lhs_symbol.name, len(lhs_symbol.arguments))
        rhs_symbol = rhs.atom.symbol
        rhs_pred = Predicate(rhs_symbol.name, len(rhs_symbol.arguments))
        # check for equality
        if lhs_pred == rhs_pred:
            for lhs_arg, rhs_arg in zip(lhs_symbol.arguments, rhs_symbol.arguments):
                if rhs_arg.ast_type == ASTType.Variable and rhs_arg.name == "_":
                    continue
                if lhs_arg != rhs_arg:
                    return False
            return True

        # check for derived superseeding
        for m in self.superseeds:
            if m.head_pred == lhs_pred and m.body_pred.pred == rhs_pred and m.body_pred.sign == rhs.sign:
                fits = True
                for rhs_index, lhs_index in enumerate(m.var_map):
                    if rhs_symbol.arguments[rhs_index] != lhs_symbol.arguments[lhs_index]:
                        fits = False
                if fits:
                    return True
        return False

    def _remove_superseed_from_list(self, body: list[AST]) -> bool:
        fix = False
        updated = False
        while not fix:
            fix = True
            for lhs, rhs in permutations(body, 2):
                if self._superseeded(lhs, rhs):
                    log.info(f"Remove {rhs} from rule, as it is superseeded by {lhs}.")
                    body.remove(rhs)
                    updated = True
                    fix = False
                    break
        return updated

    def _apply_superseeding(self, stm: AST) -> AST:
        if stm.ast_type in (ASTType.Rule, ASTType.Minimize):
            body: list[AST] = list(stm.body)
            updated = self._remove_superseed_from_list(body)
            for idx, blit in enumerate(body):
                if blit.ast_type == ASTType.ConditionalLiteral:
                    condition: list[AST] = list(blit.condition)
                    updated |= self._remove_superseed_from_list(condition)
                    body[idx] = blit.update(condition=condition)
                elif blit.ast_type == ASTType.Literal and blit.atom.ast_type in (
                    ASTType.BodyAggregate,
                    ASTType.Aggregate,
                ):
                    for element_idx, element in enumerate(blit.atom.elements):
                        condition = list(element.condition)
                        updated |= self._remove_superseed_from_list(condition)
                        blit.atom.elements[element_idx] = element.update(condition=condition)

            if updated:
                return stm.update(body=body)
        return stm

    @staticmethod
    def true(stm: AST) -> bool:
        """check wether literal is constant true"""
        if stm.ast_type == ASTType.Literal and stm.atom.ast_type == ASTType.BooleanConstant:
            if stm.sign in (Sign.NoSign, Sign.DoubleNegation):
                return bool(stm.atom.value)
            if stm.sign == Sign.Negation:  # nocoverage
                return not bool(stm.atom.value)  # nocoverage # parser turns not false into true
        if stm.ast_type == ASTType.ConditionalLiteral and not stm.condition:
            return CleanupTranslator.true(stm.literal)
        return False

    @staticmethod
    def false(stm: AST) -> bool:
        """check wether literal is constant false"""
        if stm.ast_type == ASTType.Literal and stm.atom.ast_type == ASTType.BooleanConstant:
            if stm.sign in (Sign.NoSign, Sign.DoubleNegation):
                return not bool(stm.atom.value)
            if stm.sign == Sign.Negation:  # nocoverage
                return bool(stm.atom.value)  # nocoverage # parser turns not false into true
        if stm.ast_type == ASTType.ConditionalLiteral and not stm.condition:
            return CleanupTranslator.false(stm.literal)
        return False

    @staticmethod
    def remove_true_literals(lits: Iterable[AST]) -> list[AST]:
        """remove all #true literals from the list"""
        newbody: list[AST] = []
        for lit in lits:
            if not CleanupTranslator.true(lit):
                newbody.append(lit)
        return newbody

    @staticmethod
    def contains_false(lits: Iterable[AST]) -> bool:
        """true if lits contain #false"""
        for lit in lits:
            if CleanupTranslator.false(lit):
                return True
        return False

    @staticmethod
    def cleanup_boolean_conditionals(lits: Iterable[AST]) -> list[AST]:
        """remove #true and #false from conditional literals and replace if needed"""
        ret: list[AST] = []
        for lit in lits:
            if lit.ast_type == ASTType.ConditionalLiteral:
                cond = CleanupTranslator.remove_true_literals(lit.condition)
                if not CleanupTranslator.contains_false(cond):
                    ret.append(lit.update(condition=cond))
            else:
                ret.append(lit)
        return ret

    @staticmethod
    def cleanup_boolean_aggregates(lits: Iterable[AST]) -> list[AST]:
        """remove #true and #false from BodyAggregate literals and replace if needed"""
        ret: list[AST] = []
        for lit in lits:
            if lit.ast_type == ASTType.Literal and lit.atom.ast_type == ASTType.BodyAggregate:
                new_elements: list[AST] = []
                for elem in lit.atom.elements:
                    cond = CleanupTranslator.remove_true_literals(elem.condition)
                    if not CleanupTranslator.contains_false(cond):
                        new_elements.append(elem.update(condition=cond))
                ret.append(lit.update(atom=lit.atom.update(elements=new_elements)))
            else:
                ret.append(lit)
        return ret

    def remove_boolean(self, stm: AST) -> Optional[AST]:
        """remove any constant booleans from the statement or the statement itself"""
        if stm.ast_type in (ASTType.Rule, ASTType.Minimize):
            stm = stm.update(body=self.cleanup_boolean_aggregates(stm.body))
            stm = stm.update(body=self.cleanup_boolean_conditionals(stm.body))
            stm = stm.update(body=self.remove_true_literals(stm.body))
            if self.contains_false(stm.body):
                return None
        return stm

    def execute(self, prg: list[AST]) -> list[AST]:
        """
        remove all literals that are weaker than another one from bodies/conditionals
        """
        prg = inline_arithmetic(prg)
        self._find_superseeded(prg)

        new_prg: list[AST] = []
        for stm in prg:
            r = self.remove_boolean(self._apply_superseeding(stm))
            if r:
                new_prg.append(r)

        return new_prg



import os

import nox

nox.options.sessions = "lint_ruff", "lint_pylint", "doc", "typecheck", "test"

EDITABLE_TESTS = True
PYTHON_VERSIONS = None
if "GITHUB_ACTIONS" in os.environ:
    PYTHON_VERSIONS = ["3.11"]
    EDITABLE_TESTS = False

EXTRA_INSTALL = []

@nox.session
def format(session):
    session.install(*EXTRA_INSTALL, "-e", ".[format]")
    check = "check" in session.posargs

    isort_args = ["--profile", "black", "src", "tests"]
    if check:
        isort_args.insert(0, "--check")
        isort_args.insert(1, "--diff")
    session.run("isort", *isort_args)

    black_args = ["src", "tests"]
    if check:
        black_args.insert(0, "--check")
        black_args.insert(1, "--diff")
    session.run("black", *black_args)


@nox.session
def doc(session):
    session.install(*EXTRA_INSTALL, "-e", ".[doc]")
    session.run("pdoc", "ngo", "-o", "docs", "--no-show-source")


@nox.session
def lint_ruff(session):
    session.install(*EXTRA_INSTALL, "-e", ".[lint_ruff]")
    session.run("ruff", "check", "src", "tests")


@nox.session
def lint_pylint(session):
    session.install(*EXTRA_INSTALL, "-e", ".[lint_pylint]")
    session.run("pylint", "ngo", "tests")


@nox.session
def typecheck(session):
    session.install(*EXTRA_INSTALL, "-e", ".[typecheck]")
    session.run("mypy", "--version")
    session.run("mypy", "--strict", "--untyped-calls-exclude=networkx", "--untyped-calls-exclude=sympy", "-p", "ngo", "-p", "tests")



@nox.session(python=PYTHON_VERSIONS)
def test(session):
    args = [".[test]"]
    if EDITABLE_TESTS:
        args.insert(0, "-e")
    
    session.install(*EXTRA_INSTALL, *args)
    session.run("coverage", "run", "-m", "pytest", *session.posargs)
    session.run("coverage", "report", "-m", "--fail-under=100")


@nox.session
def dev(session):
    session.install(*EXTRA_INSTALL, "-e", ".[dev]")



""" test translation of min/max aggregates using chaining """
import pytest
from clingo.ast import AST, parse_string

from ngo.normalize import postprocess, preprocess
from ngo.symmetry import SymmetryTranslator
from ngo.utils.ast import largest_subset


@pytest.mark.parametrize(
    "prg, converted_prg",
    (
        (
            "f(X) :- node(X), player(P1, X, Y), player(P2, X, Y), P1 != P2.",
            "#program base.\nf(X) :- node(X); player(_,X,Y); 2 <= #count { P1: player(P1,X,Y) }.",
        ),
        (
            "f(X) :- node(X), player(P1, X, Y), player(P2, X, Y), P1 < P2.",
            "#program base.\nf(X) :- node(X); player(_,X,Y); 2 <= #count { P1: player(P1,X,Y) }.",
        ),
        (
            "f(X) :- node(X), player(P1, X1, Y), player(P2, X2, Y), P1 != P2, X1 = X2.",
            "#program base.\nf(X) :- node(X); player(_,X1,Y); 2 <= #count { P1: player(P1,X1,Y) }.",
        ),
        (
            "f(X) :- node(X), player(P1, X1, Y), player(P2, X2, Y), not X1 != X2, P1 != P2.",
            "#program base.\nf(X) :- node(X); player(_,X1,Y); 2 <= #count { P1: player(P1,X1,Y) }.",
        ),
        (
            "f(X) :- node(X), player(P1, X, Y), player(P2, X, Y), not P1 = P2.",
            "#program base.\nf(X) :- node(X); player(_,X,Y); 2 <= #count { P1: player(P1,X,Y) }.",
        ),
        (
            "f(X) :- node(X), player(P1, X, Y, V1), player(P2, X, Y, V2), P1 != P2, V1 != V2.",
            "#program base.\nf(X) :- node(X); player(P1,X,Y,V1); player(P2,X,Y,V2); V1 != V2; P1 < P2.",
        ),
        (
            "f(V1) :- node(X), player(P1, X, Y, V1), player(P2, X, Y, V2), P1 != P2, V1 != V2.",
            "#program base.\nf(V1) :- node(X); player(P1,X,Y,V1); player(P2,X,Y,V2); P1 != P2; V1 != V2.",
        ),
        (
            "f(V1) :- node(X), player(P1, X, Y, V1), P1 != X.",
            "#program base.\nf(V1) :- node(X); player(P1,X,Y,V1); P1 != X.",
        ),
        (
            "f(X) :- node(X), player(P1, X, Y, V1), player(P2, X, Y, V2), P1 != V1, V1 != P2.",
            "#program base.\nf(X) :- node(X); player(P1,X,Y,V1); player(P2,X,Y,V2); P1 != V1; V1 != P2.",
        ),
        (
            "#false :- at(X,Y,T); at(X,Y,U); foo(Z,W,T); bar(Z,W,U); not U = T.",
            "#program base.\n#false :- at(X,Y,T); at(X,Y,U); foo(Z,W,T); bar(Z,W,U); not U = T.",
        ),
        (
            "#false :- at(X,Y,T); at(X,Y,U); minot(Z,W,V1); minot(Z,W,V2); not U = T; V1 != V2.",
            "#program base.\n#false :- minot(Z,W,_); 2 <= #count { V1: minot(Z,W,V1) }; at(X,Y,_);\
 2 <= #count { T: at(X,Y,T) }.",
        ),
        (
            "#false :- at(X,Y,T); at(X,Y,U); minot(Z,W,T); minot(Z,W,U); not U = T.",
            "#program base.\n#false :- at(X,Y,T); at(X,Y,U); minot(Z,W,T); minot(Z,W,U); T < U.",
        ),
        (
            "f(X) :- node(X), player(P1, X, Y), player(P2, X, Y), player(P3, X, Y), P1 != P2, P1 != P3, P2 != P3.",
            "#program base.\nf(X) :- node(X); player(_,X,Y); 3 <= #count { P1: player(P1,X,Y) }.",
        ),
        (
            ":- #count{W : match(M1,W), match(M2,W), match(M3,W), M1 != M2, M1 != M3, M2 != M3} >= 2.",
            """#program base.
__aux_1(W) :- match(_,W); 3 <= #count { M1: match(M1,W) }.
#false :- 2 <= #sum+ { 1,W: __aux_1(W) }.""",
        ),
        (
            ":- #count{X,Y : player(P1, X, Y, V1), player(P2, X, Y, V2), P1 != P2, V1 != V2} >= 2.",
            """#program base.
#false :- 2 <= #sum+ { 1,X,Y: player(P1,X,Y,V1), player(P2,X,Y,V2), V1 != V2, P1 < P2 }.""",
        ),
        (
            "#minimize {a}.",
            """#program base.
:~ . [a@0]""",
        ),
        (  # test domain replacement
            """
{player(P,X,Y) : name(P), pos(X), pos(Y)}.
f(X) :- node(X), player(P1, X, Y), player(P2, X, Y), P1 != P2.
            """,
            """#program base.
{ player(P,X,Y): name(P), pos(X), pos(Y) }.
__dom_player(P,X,Y) :- name(P); pos(X); pos(Y).
f(X) :- node(X); __dom_player(_,X,Y); 2 <= #count { P1: player(P1,X,Y) }.""",
        ),
        (
            "false :- sudoku(X,Y,M); sudoku(A,B,M); c1(Y); c1(B); r1(X); r1(A); X != A; Y != B.",
            """#program base.
false :- sudoku(X,Y,M); sudoku(A,B,M); c1(Y); c1(B); r1(X); r1(A); Y != B; A < X.""",
        ),
        (
            "false :- sudoku(X,Y,M); sudoku(A,B,M); c1(Y); c1(B); r1(X); r1(A); X != A; Y < B.",
            """#program base.
false :- sudoku(X,Y,M); sudoku(A,B,M); c1(Y); c1(B); r1(X); r1(A); X != A; Y < B.""",
        ),
        (
            "{ before(T1,T2,M) } :- sequence(T1,M,_); sequence(T2,M,_); T1 < T2.",
            """#program base.
{ before(T1,T2,M) } :- sequence(T1,M,_); sequence(T2,M,_); T1 < T2.""",
        ),
        (
            ":~ node(X), player(P1, X, Y), player(P2, X, Y), P1 != P2. [X@1]",
            "#program base.\n:~ node(X); player(_,X,Y); 2 <= #count { P1: player(P1,X,Y) }. [X@1]",
        ),
        (
            ":~ node(X), player(P1, X, Y), player(P2, X, Y), P1 != P2. [P1@1]",
            "#program base.\n:~ node(X); player(P1,X,Y); player(P2,X,Y); P1 != P2. [P1@1]",
        ),
        (
            ":~ #count{W : match(M1,W), match(M2,W), match(M3,W), M1 != M2, M1 != M3, M2 != M3} >= 2. [3@1]",
            """#program base.
__aux_1(W) :- match(_,W); 3 <= #count { M1: match(M1,W) }.
:~ 2 <= #sum+ { 1,W: __aux_1(W) }. [3@1]""",
        ),
        (
            "#false :- not positive(C1); not positive(C2); edge(P,C1); edge(P,C2); C1 != C2; leaf(C1); leaf(C2).",
            """#program base.
#false :- not positive(C1); not positive(C2); edge(P,C1); edge(P,C2); leaf(C1); leaf(C2); C1 < C2.""",
        ),
        (
            "#false :- edge(X,Y); edge(X,U); edge(A,B); edge(A,M); Y < U; B < M; X < A; not U < B.",
            """#program base.
#false :- edge(X,Y); edge(X,U); edge(A,B); edge(A,M); Y < U; B < M; X < A; not U < B.""",
        ),
        (
            """
  node((1..n)).
  root(1).
  { edge(X,Y): node(Y) } :- node(X).
  1 = { decision(X,A): attribute(A) } :- node(X); not leaf(X).
  attribute(A) :- holds(_,A).
  leaf(N) :- node(N); not edge(N,_).
  { positive(N) } :- leaf(N).
  path(X,E) :- example(E,_); root(X).
  path(C1,E) :- path(X_P,E); decision(X_P,A); holds(E,A); edge(X_P,C1); edge(X_P,C2); C1 > C2. 
  path(C2,E) :- path(X_P,E); decision(X_P,A); not holds(E,A); edge(X_P,C1); edge(X_P,C2); C1 > C2.
  #false :- path(C1,E); path(C2,E); decision(X,A); decision(Y,A); X != Y; C1 != C2.""",
            """#program base.
node((1..n)).
root(1).
{ edge(X,Y): node(Y) } :- node(X).
1 = { decision(X,A): attribute(A) } :- node(X); not leaf(X).
attribute(A) :- holds(_,A).
leaf(N) :- node(N); not edge(N,_).
{ positive(N) } :- leaf(N).
path(X,E) :- example(E,_); root(X).
path(C1,E) :- path(X_P,E); decision(X_P,A); holds(E,A); edge(X_P,C1); edge(X_P,C2); C1 > C2.
path(C2,E) :- path(X_P,E); decision(X_P,A); not holds(E,A); edge(X_P,C1); edge(X_P,C2); C1 > C2.
__dom_edge(X,Y) :- node(Y); node(X).
__dom_leaf(N) :- node(N); not __dom_edge(N,_).
__dom_decision(X,A) :- attribute(A); node(X); not __dom_leaf(X).
#false :- __dom_decision(_,A); 2 <= #count { X: decision(X,A) }; path(_,E); 2 <= #count { C1: path(C1,E) }.""",
        ),
    ),
)
def test_symmetry(prg: str, converted_prg: str) -> None:
    """test symmetry breaking on whole programs"""
    ast: list[AST] = []
    parse_string(prg, ast.append)
    ast = preprocess(ast)
    st = SymmetryTranslator(ast, [])
    ast = st.execute(ast)
    ast = postprocess(ast)
    output = "\n".join(map(str, ast))
    assert converted_prg == output


@pytest.mark.parametrize(
    "input_, output",
    [
        (
            (1, 2, 3, 4),
            [
                (1, 2, 3, 4),
                (2, 3, 4),
                (1, 3, 4),
                (1, 2, 4),
                (1, 2, 3),
                (3, 4),
                (2, 4),
                (2, 3),
                (1, 4),
                (1, 3),
                (1, 2),
                (4,),
                (3,),
                (2,),
                (1,),
                tuple(),
            ],
        )
    ],
)
def test_largest_subset(input_: tuple[int, ...], output: list[tuple[int, ...]]) -> None:
    """ "test largest subset order"""
    assert list(largest_subset(input_)) == output



""" test ast utility functions """
import pytest
from clingo.ast import AST, parse_string

from ngo.normalize import postprocess, preprocess


@pytest.mark.parametrize(
    "input_, output",
    [
        (
            """
a :- b(X); X = #sum {Y : Y=1..X}.
""",
            """#program base.
a :- b(X); X = #sum { Y: Y = (1..X) }.""",
        ),
        (
            """
#false :- sudoku(X,_,N); 1 != { sudoku(X,_,N) }.
""",
            """#program base.
#false :- sudoku(X,_,N); 1 != #sum { 1,0,sudoku(X,AUX,N): sudoku(X,AUX,N) }.""",
        ),
        (
            """
#false :- sudoku(X,_,N); 1 != { not sudoku(X,_,N) }.
""",
            """#program base.
#false :- sudoku(X,_,N); 1 != #sum { 1,1,sudoku(X,anon__ngo,N): not sudoku(X,_,N) }.""",
        ),
        (
            """
#false :- N != { bend(((T+1)..(T1-1))) }; bends(T,T1,N).
""",
            """#program base.
#false :- N != #sum { 1,0,bend(AUX): bend(AUX), AUX = ((T+1)..(T1-1)) }; bends(T,T1,N).""",
        ),
    ],
)
def test_preprocess(input_: str, output: str) -> None:
    """test preprocessing"""
    ast: list[AST] = []
    parse_string(input_, ast.append)
    ast = preprocess(ast)
    assert str("\n".join(map(str, ast))) == output


@pytest.mark.parametrize(
    "input_, output",
    [
        (
            """
a :- b(X); X = #sum {Y : Y=1..X}.
""",
            """#program base.
a :- b(X); X = #sum { Y: Y = (1..X) }.""",
        ),
        (
            """
a :- b(X); X = #sum {Y : f(Y), Y=1..17}.
""",
            """#program base.
a :- b(X); X = #sum { Y: f(Y), Y = (1..17) }.""",
        ),
        (
            """
a :- b(X); X = #sum {Y : f(Y), Y=3+4}.
""",
            """#program base.
a :- b(X); X = #sum { (3+4): f((3+4)) }.""",
        ),
        (
            """
a :- b(X); X = #sum {Y : f(Y), Y=3+4; 15}.
""",
            """#program base.
a :- b(X); X = #sum { (3+4): f((3+4)); 15 }.""",
        ),
        (
            """
:~ a(X); b(Y); X = (Y+3). [X@1]
""",
            """#program base.
:~ a((Y+3)); b(Y). [(Y+3)@1]""",
        ),
        (
            """
a :- b(X): c(Y), Y = (4+7).
""",
            """#program base.
a :- b(X): c((4+7)).""",
        ),
        (
            """
a :- b(X): c(X), _ = (4+7).
""",
            """#program base.
a :- b(X): c(X), _ = (4+7).""",
        ),
        (
            """
a :- b(X): c(X), (4+7) = _.
""",
            """#program base.
a :- b(X): c(X), (4+7) = _.""",
        ),
    ],
)
def test_postprocess(input_: str, output: str) -> None:
    """test preprocessing"""
    ast: list[AST] = []
    parse_string(input_, ast.append)
    ast = postprocess(ast)
    assert str("\n".join(map(str, ast))) == output
